"""
RETURNN frontend (returnn.frontend) tests
"""

from __future__ import annotations
from typing import Tuple
from unittest import SkipTest
import numpy
import numpy.testing
import _setup_test_env  # noqa
import returnn.frontend as rf
from returnn.tensor import Tensor, Dim, TensorDict, batch_dim
from returnn.tensor.utils import tensor_dict_fill_random_numpy_
from rf_utils import run_model, run_model_torch_train


# Keep test_linear_direct and test_linear first here to have some very canonical examples.


def _setup():
    rf.select_backend_torch()  # enables some of the native optimizations


_setup()


def test_linear_direct():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim, out_dim = Dim(7, name="in"), Dim(13, name="out")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
            "classes": Tensor("classes", [batch_dim, time_dim], dtype="int32", sparse_dim=out_dim),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, model: rf.Linear, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output()

    run_model(extern_data, lambda *, epoch, step: rf.Linear(in_dim, out_dim), _forward_step)


def test_linear():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim, out_dim = Dim(7, name="in"), Dim(13, name="out")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
            "classes": Tensor("classes", [batch_dim, time_dim], dtype="int32", sparse_dim=out_dim),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.linear = rf.Linear(in_dim, out_dim)

        def __call__(self, x: Tensor) -> Tensor:
            return self.linear(x)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output()

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


# Now come some tests for some base functionality.


def test_train_flag():
    rf.init_train_step_run_ctx(train_flag=False)
    assert rf.get_run_ctx().train_flag is False
    assert rf.get_run_ctx().is_train_flag_enabled(func=rf.dropout) is False
    with rf.get_run_ctx().train_flag_ctx(True):
        assert rf.get_run_ctx().train_flag is True
        assert rf.get_run_ctx().is_train_flag_enabled(func=rf.dropout) is True
        with rf.get_run_ctx().train_flag_ctx(False):
            assert rf.get_run_ctx().train_flag is False
            assert rf.get_run_ctx().is_train_flag_enabled(func=rf.dropout) is False
            with rf.get_run_ctx().train_flag_ctx(True, func=rf.dropout):
                assert rf.get_run_ctx().train_flag is False
                assert rf.get_run_ctx().is_train_flag_enabled(func=rf.dropout) is True


def test_state():
    # https://github.com/rwth-i6/returnn/issues/1329
    import tree

    s = rf.LstmState(h=Tensor("h", (), "float32"), c=Tensor("c", (), "float32"))
    res = tree.map_structure(lambda x: x, s)
    assert isinstance(res, rf.LstmState)
    assert res is not s
    assert res.h is s.h and res.c is s.c


# And now more module tests.


def test_2layers():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim, hidden_dim, out_dim = Dim(7, name="in"), Dim(11, name="hidden"), Dim(13, name="out")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
            "classes": Tensor("classes", [batch_dim, time_dim], dtype="int32", sparse_dim=out_dim),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.layer1 = rf.Linear(in_dim, hidden_dim)
            self.layer2 = rf.Linear(hidden_dim, out_dim)

        def __call__(self, x: Tensor) -> Tensor:
            x = rf.relu(self.layer1(x))
            x = self.layer2(x)
            return x

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output()

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_linear_same_dim():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim, out_dim = Dim(7, name="in"), Dim(13, name="out")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
            "classes": Tensor("classes", [batch_dim, time_dim], dtype="int32", sparse_dim=out_dim),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.layer1 = rf.Linear(in_dim, out_dim)
            self.layer2 = rf.Linear(out_dim, out_dim)

        def __call__(self, x: Tensor) -> Tensor:
            x = rf.relu(self.layer1(x))
            x = self.layer2(x)
            return x

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output()

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_linear_cross_entropy():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim, out_dim = Dim(7, name="in"), Dim(13, name="out")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
            "classes": Tensor("classes", [batch_dim, time_dim], dtype="int32", sparse_dim=out_dim),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.layer1 = rf.Linear(in_dim, out_dim)
            self.layer2 = rf.Linear(out_dim, out_dim)

        def __call__(self, x: Tensor) -> Tensor:
            x = rf.relu(self.layer1(x))
            x = self.layer2(x)
            return x

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        logits = model(extern_data["data"])
        targets = extern_data["classes"]
        loss = rf.cross_entropy(estimated=logits, estimated_type="logits", target=targets, axis=out_dim)
        loss.mark_as_default_output()

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_linear_ctc():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    target_time_dim = Dim(Tensor("target_time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    hidden_dim = Dim(13, name="hidden")
    out_dim = Dim(11, name="classes")
    out_wb_dim = out_dim + 1
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
            "classes": Tensor("classes", [batch_dim, target_time_dim], dtype="int32", sparse_dim=out_dim),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.layer1 = rf.Linear(in_dim, hidden_dim)
            self.layer2 = rf.Linear(hidden_dim, out_wb_dim)

        def __call__(self, x: Tensor) -> Tensor:
            x = rf.relu(self.layer1(x))
            x = self.layer2(x)
            return x

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        logits = model(extern_data["data"])
        targets = extern_data["classes"]
        loss = rf.ctc_loss(
            logits=logits,
            targets=targets,
            input_spatial_dim=time_dim,
            targets_spatial_dim=target_time_dim,
            blank_index=out_wb_dim.dimension - 1,
        )
        loss.mark_as_default_output()

    run_model(
        extern_data,
        lambda *, epoch, step: _Net(),
        _forward_step,
        dyn_dim_min_sizes={time_dim: 4, target_time_dim: 2},
        dyn_dim_max_sizes={time_dim: 11, target_time_dim: 5},
    )


def test_dropout():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            return rf.dropout(x, 0.5, axis=rf.dropout_broadcast_default() and in_dim, on_forward=True)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_dim_value():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            res = rf.ones((), dtype="int64")
            for d in x.dims:
                res *= rf.cast(rf.convert_to_tensor(d.get_dim_value_tensor()), "int64")
            return res

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=())

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_dim_size_after_declare_same_as():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim], dtype="float32"),
        }
    )
    out_spatial_dim = Dim(None, name="out_spatial")

    # noinspection PyShadowingNames
    def _forward_step(**_kwargs):
        assert time_dim.dyn_size_ext.raw_tensor is not None
        enc_spatial_dim = time_dim * 2
        assert enc_spatial_dim.dyn_size_ext.raw_tensor is not None
        out_spatial_dim.declare_same_as(enc_spatial_dim)
        assert enc_spatial_dim.dyn_size_ext.raw_tensor is not None
        assert out_spatial_dim.dyn_size_ext.raw_tensor is not None
        out_spatial_dim.dyn_size_ext.mark_as_default_output(shape=[batch_dim])

    # Running it twice can trigger some weird behavior in Dim.declare_same_as.
    run_model(extern_data, lambda **_kwargs: rf.Module(), _forward_step, test_tensorflow=False)
    run_model(extern_data, lambda **_kwargs: rf.Module(), _forward_step, test_tensorflow=False)


def test_dim_mask():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            mask1 = time_dim.get_mask()
            mask1.verify_out_shape({batch_dim, time_dim})
            mask2 = time_dim.get_mask()
            assert mask1 is mask2  # cache
            time_dim_copy = Dim(None, name="time_copy")
            time_dim_copy.copy_from(time_dim)
            assert time_dim_copy != time_dim
            mask3 = time_dim_copy.get_mask()
            mask3.verify_out_shape({batch_dim, time_dim_copy})
            assert mask1 is not mask3  # other dims
            if rf.is_backend_raw_tensor_dim_tag_independent():
                assert mask1.raw_tensor is mask3.raw_tensor  # cache even across copy
            return mask1

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_num_elements_of_shape():
    import numpy as np

    batch_dim_ = Dim(5, name="batch")
    enc_dim = Dim(Tensor("enc", dims=[batch_dim_], dtype="int64"))
    dec_dim = Dim(Tensor("dec", dims=[batch_dim_], dtype="int64"))
    enc_dim.dyn_size_ext.raw_tensor = np.array([17, 16, 15, 13, 12])
    dec_dim.dyn_size_ext.raw_tensor = np.array([11, 10, 8, 7, 5])
    calc_n_enc = sum(enc_dim.dyn_size_ext.raw_tensor)
    calc_n_dec = sum(dec_dim.dyn_size_ext.raw_tensor)
    calc_n_prod = sum(enc_dim.dyn_size_ext.raw_tensor * dec_dim.dyn_size_ext.raw_tensor)
    assert rf.num_elements_of_shape([batch_dim_]) == batch_dim_.dimension
    n_b_enc = rf.num_elements_of_shape([batch_dim_, enc_dim])
    assert calc_n_enc == n_b_enc.raw_tensor.item()
    n_b_dec = rf.num_elements_of_shape([dec_dim, batch_dim_])
    assert calc_n_dec == n_b_dec.raw_tensor.item()
    n_prod = rf.num_elements_of_shape([batch_dim_, enc_dim, dec_dim])
    assert calc_n_prod == n_prod.raw_tensor.item()


def test_convert_to_tensor_numpy_backend():
    import numpy as np
    from returnn.frontend._numpy_backend import NumpyBackend

    x = rf.convert_to_tensor(1, dims=(), dtype="int32", _backend=NumpyBackend)
    assert isinstance(x.raw_tensor, np.ndarray)
    assert x.raw_tensor.dtype == np.int32
    assert x.raw_tensor.item() == 1


def test_param_assign():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.param = rf.Parameter(dims=(), dtype="int32")
            self.param.initial = 2

        def __call__(self, x: Tensor) -> Tuple[Tensor, Tensor, Tensor]:
            # No extra care should be needed for graph-based backends.
            a = rf.copy(self.param)
            self.param.assign(5)
            b = rf.copy(self.param)
            self.param.assign(7)
            return a, b, self.param

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        a, b, c = model(extern_data["data"])
        a.mark_as_output("a", shape=())
        b.mark_as_output("b", shape=())
        c.mark_as_output("c", shape=())

    out = run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)
    assert out["a"].raw_tensor == 2 and out["b"].raw_tensor == 5 and out["c"].raw_tensor == 7


def test_loss_normalized():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames
    def _train_step(*, model: rf.Module, extern_data: TensorDict):
        model  # unused  # noqa
        x = extern_data["data"]

        loss = rf.reduce_sum(x, axis=in_dim)  # [B,T]
        loss.mark_as_loss("loss", use_normalized_loss=True)

        loss_custom_norm = rf.reduce_sum(loss, axis=time_dim)  # [B]
        loss_custom_norm.mark_as_loss(
            "loss_custom_norm", custom_inv_norm_factor=time_dim.get_size_tensor(), use_normalized_loss=True
        )

    run_model_torch_train(extern_data, lambda *, epoch, step: rf.Module(), _train_step)


def test_loss_normalization():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    use_normalized = False
    use_custom_inv_norm_factor = False

    # noinspection PyShadowingNames
    def _train_step(*, model: rf.Module, extern_data: TensorDict):
        model  # unused  # noqa
        x = extern_data["data"]

        loss = rf.reduce_sum(x, axis=in_dim)  # [B,T]
        loss.mark_as_loss(
            "loss",
            custom_inv_norm_factor=time_dim.get_size_tensor() if use_custom_inv_norm_factor else None,
            use_normalized_loss=use_normalized,
        )

    res1 = run_model_torch_train(extern_data, lambda *, epoch, step: rf.Module(), _train_step)
    res2 = run_model_torch_train(extern_data, lambda *, epoch, step: rf.Module(), _train_step)
    assert res1 == res2  # check deterministic
    use_normalized = True
    res3 = run_model_torch_train(extern_data, lambda *, epoch, step: rf.Module(), _train_step)
    assert res3["loss:summed"] == res2["loss:summed"] and res3["loss:inv_norm_factor"] == res2["loss:inv_norm_factor"]
    use_custom_inv_norm_factor = True
    res4 = run_model_torch_train(extern_data, lambda *, epoch, step: rf.Module(), _train_step)
    assert res4["loss:summed"] == res2["loss:summed"] and res4["loss:inv_norm_factor"] == res2["loss:inv_norm_factor"]


def test_rf_range_over_dim():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, model: rf.Module, extern_data: TensorDict):
        rf.range_over_dim(time_dim).mark_as_output("range", shape=[time_dim])
        rf.range_over_dim(time_dim, dtype="float32").mark_as_output("range_float", shape=[time_dim])

    run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step)


def test_build_dict():
    d = rf.build_dict(rf.BatchNorm, use_mask=True)
    assert d == {"class": "rf.BatchNorm", "use_mask": True}


def test_build_dict_func():
    d = rf.build_dict(rf.relu)
    assert d == {"class": "rf.relu"}


def test_build_from_dict():
    rf.select_backend_torch()
    in_dim = Dim(7, name="in")
    mod = rf.build_from_dict({"class": "rf.BatchNorm", "use_mask": True}, in_dim, eps=1e-4)
    assert isinstance(mod, rf.BatchNorm)
    assert mod.in_dim == in_dim
    assert mod.use_mask
    assert mod.eps == 1e-4


def test_build_from_dict_func():
    import functools

    func = rf.build_from_dict({"class": "rf.relu"})
    assert func is rf.relu
    func = rf.build_from_dict({"class": "rf.combine", "kind": "+", "b": 1})
    assert isinstance(func, functools.partial)
    assert func.func is rf.combine
    assert func.keywords == {"kind": "+", "b": 1}


def test_build_from_dict_func_native():
    from types import BuiltinFunctionType

    rf.select_backend_torch()  # enables some of the native optimizations
    assert isinstance(rf.combine, BuiltinFunctionType)  # due to native optimizations
    func = rf.build_from_dict({"class": "rf.combine"})
    assert func is rf.combine


def test_parametrization():
    from functools import partial

    rf.select_backend_torch()  # any, doesn't really matter for the test
    rf.init_train_step_run_ctx(train_flag=True)  # such that dropout is used below

    in_dim = Dim(7, name="in")
    out_dim = Dim(13, name="out")
    mod = rf.Linear(in_dim, out_dim)
    orig_weight = mod.weight
    assert isinstance(orig_weight, rf.Parameter)
    orig_bias = mod.bias

    # Test parametrization.
    rf.register_parametrization(mod, "weight", partial(rf.dropout, drop_prob=0.5))
    assert rf.is_parametrized(mod)
    assert rf.is_parametrized(mod, "weight")
    weight = mod.weight
    assert weight is not orig_weight and not isinstance(weight, rf.Parameter)
    params = dict(mod.named_parameters())
    assert set(params.keys()) == {"weight", "bias"}
    assert params["weight"] is orig_weight
    assert params["bias"] is orig_bias

    rf.init_train_step_run_ctx(train_flag=False)
    weight = mod.weight
    assert weight is orig_weight  # no dropout in eval mode

    rf.init_train_step_run_ctx(train_flag=True)  # such that dropout would be used again
    rf.remove_parametrization(mod, "weight")
    weight = mod.weight
    assert weight is orig_weight
    assert not rf.is_parametrized(mod, "weight")
    assert not rf.is_parametrized(mod)
    params = dict(mod.named_parameters())
    assert set(params.keys()) == {"weight", "bias"}
    assert params["weight"] is orig_weight
    assert params["bias"] is orig_bias


def test_weight_noise():
    import torch

    if torch.__version__ < (2, 0):
        raise SkipTest("Torch version too old for this test (gradient_checkpoint_scope needs Torch >= 2.0)")
    rf.select_backend_torch()  # any, doesn't really matter for the test
    rf.init_train_step_run_ctx(train_flag=True)  # such that weight noise is used below

    in_dim = Dim(7, name="in")
    out_dim = Dim(13, name="out")
    mod = rf.Linear(in_dim, out_dim)
    orig_weight = mod.weight
    assert isinstance(orig_weight, rf.Parameter)
    orig_bias = mod.bias

    # Test parametrization.
    rf.weight_noise(mod, "weight", std=0.1)
    assert rf.is_parametrized(mod, "weight")
    weight = mod.weight
    assert weight is not orig_weight and not isinstance(weight, rf.Parameter)
    params = dict(mod.named_parameters())
    assert set(params.keys()) == {"weight", "bias"}
    assert params["weight"] is orig_weight
    assert params["bias"] is orig_bias

    class _Model(rf.Module):
        def __init__(self):
            super().__init__()
            self.linear = rf.Linear(in_dim, out_dim)

    model = _Model()
    for mod in model.modules():
        for param_name, param in mod.named_parameters(recurse=False):
            if param_name.endswith("bias"):  # no bias
                continue
            print("***", mod, param_name, param)
            rf.weight_noise(mod, param_name, std=0.0025)

    # https://github.com/rwth-i6/returnn/issues/1580
    conv = rf.Conv1d(in_dim, out_dim, 3, padding="same")
    rf.weight_noise(conv, "filter", std=0.1)
    time_dim = Dim(11, name="time")
    conv(rf.random_normal([time_dim, in_dim]), in_spatial_dim=time_dim)


def test_ctc_loss():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    target_time_dim = Dim(Tensor("target_time", [batch_dim], dtype="int32"))
    out_dim = Dim(11, name="classes")
    out_wb_dim = out_dim + 1
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, out_wb_dim], dtype="float32", feature_dim=out_wb_dim),
            "classes": Tensor("classes", [batch_dim, target_time_dim], dtype="int32", sparse_dim=out_dim),
        }
    )

    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        logits = extern_data["data"]
        targets = extern_data["classes"]
        loss = rf.ctc_loss(
            logits=logits,
            targets=targets,
            input_spatial_dim=time_dim,
            targets_spatial_dim=target_time_dim,
            blank_index=out_wb_dim.dimension - 1,
        )
        loss.mark_as_default_output()

    run_model(
        extern_data,
        lambda **_kwargs: rf.Module(),
        _forward_step,
        dyn_dim_min_sizes={time_dim: 4, target_time_dim: 2},
        dyn_dim_max_sizes={time_dim: 11, target_time_dim: 5},
    )


def test_ctc_loss_native_vs_std():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    target_time_dim = Dim(Tensor("target_time", [batch_dim], dtype="int32"))
    out_dim = Dim(11, name="classes")
    out_wb_dim = out_dim + 1
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, out_wb_dim], dtype="float32", feature_dim=out_wb_dim),
            "classes": Tensor("classes", [batch_dim, target_time_dim], dtype="int32", sparse_dim=out_dim),
        }
    )

    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        logits = extern_data["data"]
        targets = extern_data["classes"]
        loss_native = rf.ctc_loss(
            logits=logits,
            targets=targets,
            input_spatial_dim=time_dim,
            targets_spatial_dim=target_time_dim,
            blank_index=out_wb_dim.dimension - 1,
            use_native_op=True,
        )
        loss_native.mark_as_output("native", shape=[batch_dim])
        loss_std = rf.ctc_loss(
            logits=logits,
            targets=targets,
            input_spatial_dim=time_dim,
            targets_spatial_dim=target_time_dim,
            blank_index=out_wb_dim.dimension - 1,
            use_native_op=False,
        )
        loss_std.mark_as_output("std", shape=[batch_dim])

    res = run_model(
        extern_data,
        lambda **_kwargs: rf.Module(),
        _forward_step,
        dyn_dim_min_sizes={time_dim: 4, target_time_dim: 2},
        dyn_dim_max_sizes={time_dim: 11, target_time_dim: 5},
        test_tensorflow=False,
    )
    # test equality
    res_native = res["native"].raw_tensor
    res_std = res["std"].raw_tensor
    assert res_native.shape == res_std.shape
    numpy.testing.assert_allclose(res_native, res_std, rtol=1e-5, atol=1e-6)


def test_ctc_loss_broadcast():
    branch_dim = Dim(3, name="branch")
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    target_time_dim = Dim(Tensor("target_time", [batch_dim], dtype="int32"))
    out_dim = Dim(11, name="classes")
    out_wb_dim = out_dim + 1
    extern_data = TensorDict(
        {
            "data": Tensor(
                "data", [batch_dim, branch_dim, time_dim, out_wb_dim], dtype="float32", feature_dim=out_wb_dim
            ),
            "classes": Tensor("classes", [batch_dim, target_time_dim], dtype="int32", sparse_dim=out_dim),
        }
    )

    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        logits = extern_data["data"]
        targets = extern_data["classes"]
        loss = rf.ctc_loss(
            logits=logits,
            targets=targets,
            input_spatial_dim=time_dim,
            targets_spatial_dim=target_time_dim,
            blank_index=out_wb_dim.dimension - 1,
        )
        loss.mark_as_default_output()

    run_model(
        extern_data,
        lambda **_kwargs: rf.Module(),
        _forward_step,
        test_tensorflow=False,
        dyn_dim_min_sizes={time_dim: 4, target_time_dim: 2},
        dyn_dim_max_sizes={time_dim: 11, target_time_dim: 5},
    )


def test_ctc_best_path(*, label_loop: bool = True):
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    target_time_dim = Dim(Tensor("target_time", [batch_dim], dtype="int32"))
    out_dim = Dim(11, name="classes")
    out_wb_dim = out_dim + 1
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, out_wb_dim], dtype="float32", feature_dim=out_wb_dim),
            "classes": Tensor("classes", [batch_dim, target_time_dim], dtype="int32", sparse_dim=out_dim),
        }
    )

    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        logits = extern_data["data"]
        targets = extern_data["classes"]
        log_probs = rf.log_softmax(logits, axis=out_wb_dim)
        log_probs.mark_as_output("log_probs", shape=[time_dim, batch_dim, out_wb_dim])
        targets.mark_as_output("targets", shape=[batch_dim, target_time_dim])
        alignment = rf.ctc_best_path(
            logits=log_probs,
            logits_normalized=True,
            targets=targets,
            input_spatial_dim=time_dim,
            targets_spatial_dim=target_time_dim,
            blank_index=out_wb_dim.dimension - 1,
            label_loop=label_loop,
        )
        alignment.mark_as_default_output(shape=[time_dim, batch_dim])

    res = run_model(
        extern_data,
        lambda **_kwargs: rf.Module(),
        _forward_step,
        dyn_dim_min_sizes={time_dim: 4, target_time_dim: 2},
        dyn_dim_max_sizes={time_dim: 11, target_time_dim: 5},
        test_tensorflow=False,
    )

    from fsa_utils import py_viterbi
    from returnn.util.fsa import get_ctc_fsa_fast_bw as get_ctc_fsa_fast_bw_np

    # test equality
    targets = res["targets"].raw_tensor.astype("int32")
    target_seq_lens = res["targets"].dims[1].dyn_size.astype("int32")
    blank_idx = out_wb_dim.dimension - 1
    fsa = get_ctc_fsa_fast_bw_np(targets=targets, seq_lens=target_seq_lens, blank_idx=blank_idx, label_loop=label_loop)
    assert fsa.start_end_states.shape == (2, len(target_seq_lens))
    edges = fsa.edges.astype("int32")
    weights = fsa.weights.astype("float32")
    start_end_states = fsa.start_end_states.astype("int32")
    log_probs = res["log_probs"].raw_tensor.astype("float32")
    time_seq_lens = res["log_probs"].dims[0].dyn_size.astype("int32")
    ref_alignment, _ = py_viterbi(
        am_scores=log_probs, am_seq_len=time_seq_lens, edges=edges, weights=weights, start_end_states=start_end_states
    )
    res_alignment = res["output"].raw_tensor
    assert ref_alignment.shape == res_alignment.shape
    numpy.testing.assert_equal(ref_alignment, res_alignment)


def test_ctc_best_path_rna():
    test_ctc_best_path(label_loop=False)


def test_ctc_durations_from_path():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    out_dim = Dim(11, name="classes")
    out_wb_dim = out_dim + 1
    b = out_wb_dim.dimension - 1
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim], dtype="int32", sparse_dim=out_wb_dim),
            "blanks": Tensor("blanks", [batch_dim, time_dim], dtype="bool"),  # more blanks
        }
    )

    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        path = extern_data["data"]
        path = rf.where(extern_data["blanks"], b, path)
        durations, dur_spatial_dim = rf.ctc_durations_from_path(path=path, path_spatial_dim=time_dim, blank_index=b)
        durations.mark_as_default_output(shape=[batch_dim, dur_spatial_dim])

    # Now reset the extern_data manually
    extern_data.reset_content()
    tensor_dict_fill_random_numpy_(
        extern_data,
        dyn_dim_min_sizes={batch_dim: 10, time_dim: 15},
        dyn_dim_max_sizes={batch_dim: 10, time_dim: 20},
        rnd=42,
    )
    examples = [
        ([b, b, 1, 1, 2, b, b, 3, 3, 3, b], [2, 2, 0, 1, 2, 3, 1]),  # [_ _ a a b _ _ c c c _]
        ([b, b, 1, 1, b, 2, b, b, b, b], [2, 2, 1, 1, 4]),
        ([1, 1, 1, b, b, b, b], [0, 3, 4]),
        ([b, b, b, b, b, 1, 2, 3], [5, 1, 0, 1, 0, 1, 0]),
        ([5], [0, 1, 0]),
    ]
    for i, (example, _) in enumerate(examples):
        extern_data["data"].raw_tensor[i, : len(example)] = numpy.array(example, dtype="int32")
        time_dim.dyn_size_ext.raw_tensor[i] = len(example)
        extern_data["blanks"].raw_tensor[i] = False  # keep our example as is
    # Make sure at least one of the dyn sizes matches the max size.
    time_dim.dyn_size_ext.raw_tensor[len(examples)] = time_dim.get_dim_value()

    res = run_model(
        extern_data,
        lambda **_kwargs: rf.Module(),
        _forward_step,
        data_fill_random=False,
        test_tensorflow=False,  # no scatter
    )

    output = res["output"]
    out_spatial_dim = output.dims[1]
    for i, (_, exp) in enumerate(examples):
        assert out_spatial_dim.dyn_size_ext.raw_tensor[i] == len(exp)
        dur = res["output"].raw_tensor[i, : len(exp)]
        numpy.testing.assert_array_equal(dur, numpy.array(exp, dtype="int32"))


def test_ctc_no_label_loop_blank_durations_from_path():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    out_dim = Dim(11, name="classes")
    out_wb_dim = out_dim + 1
    b = out_wb_dim.dimension - 1
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim], dtype="int32", sparse_dim=out_wb_dim),
            "blanks": Tensor("blanks", [batch_dim, time_dim], dtype="bool"),  # more blanks
        }
    )

    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        path = extern_data["data"]
        path = rf.where(extern_data["blanks"], b, path)
        durations, dur_spatial_dim = rf.ctc_no_label_loop_blank_durations_from_path(
            path=path, path_spatial_dim=time_dim, blank_index=b
        )
        durations.mark_as_default_output(shape=[batch_dim, dur_spatial_dim])

    # Now reset the extern_data manually
    extern_data.reset_content()
    tensor_dict_fill_random_numpy_(
        extern_data,
        dyn_dim_min_sizes={batch_dim: 10, time_dim: 15},
        dyn_dim_max_sizes={batch_dim: 10, time_dim: 20},
        rnd=42,
    )
    examples = [
        ([b, b, b, 1, 2, b, b, 3, b], [3, 0, 2, 1]),  # [_ _ _ a b _ _ c _]
        ([b, b, 1, 1, 2, b, b, 3, b], [2, 0, 0, 2, 1]),
        ([1], [0, 0]),
        ([b, b, 1], [2, 0]),
    ]
    for i, (example, _) in enumerate(examples):
        extern_data["data"].raw_tensor[i, : len(example)] = numpy.array(example, dtype="int32")
        time_dim.dyn_size_ext.raw_tensor[i] = len(example)
        extern_data["blanks"].raw_tensor[i] = False  # keep our example as is
    # Make sure at least one of the dyn sizes matches the max size.
    time_dim.dyn_size_ext.raw_tensor[len(examples)] = time_dim.get_dim_value()

    res = run_model(
        extern_data,
        lambda **_kwargs: rf.Module(),
        _forward_step,
        data_fill_random=False,
        test_tensorflow=False,  # no scatter
    )

    output = res["output"]
    out_spatial_dim = output.dims[1]
    for i, (_, exp) in enumerate(examples):
        assert out_spatial_dim.dyn_size_ext.raw_tensor[i] == len(exp)
        dur = res["output"].raw_tensor[i, : len(exp)]
        numpy.testing.assert_array_equal(dur, numpy.array(exp, dtype="int32"))


def test_edit_distance():
    import numpy
    import torch
    from collections import namedtuple
    import itertools
    from numpy_ref_edit_distance import edit_distance_ref

    # noinspection PyShadowingNames
    def _check_edit_distance(a: Tensor, a_spatial_dim: Dim, b: Tensor, b_spatial_dim: Dim):
        ref = edit_distance_ref(a, a_spatial_dim, b, b_spatial_dim).raw_tensor
        res = rf.edit_distance(a, a_spatial_dim, b, b_spatial_dim)
        assert res.raw_tensor.shape == ref.shape == a_spatial_dim.dyn_size.shape == b_spatial_dim.dyn_size.shape
        assert len(ref.shape) == 1
        print("ref:", ref, "res:", res.raw_tensor)
        batch_size = ref.shape[0]
        for i in range(batch_size):
            assert res.raw_tensor[i] == ref[i], (
                f"batch idx i={i}, a[i]={a.raw_tensor[i]} len {a_spatial_dim.dyn_size[i]},"
                f" b[i]={b.raw_tensor[i]} len {b_spatial_dim.dyn_size[i]},"
                f" ref[i]={ref[i]}, res[i]={res.raw_tensor[i]};\n"
                f" a={a.raw_tensor} lens {a_spatial_dim.dyn_size},"
                f" b={b.raw_tensor} lens {b_spatial_dim.dyn_size}"
            )
        assert (res.raw_tensor.numpy() == ref).all()

    SizedTensor = namedtuple("SizedTensor", ["tensor", "seq_lens"])

    _SeqsB1 = [
        SizedTensor(torch.tensor([[1, 2, 3, 4]]), torch.tensor([4])),
        SizedTensor(torch.tensor([[1, 2, 3]]), torch.tensor([3])),
        SizedTensor(torch.tensor([[1, 2, 4]]), torch.tensor([3])),
        SizedTensor(torch.tensor([[1, 4]]), torch.tensor([2])),
        SizedTensor(torch.tensor([[5, 2, 4]]), torch.tensor([3])),
        SizedTensor(torch.tensor([[]], dtype=torch.int64), torch.tensor([0])),
    ]

    for a, b in itertools.product(_SeqsB1, _SeqsB1):
        a: SizedTensor
        b: SizedTensor
        # noinspection PyShadowingNames
        batch_dim = Dim(1, name="batch")
        a_spatial_dim = Dim(Tensor("a_sizes", [batch_dim], dtype="int64", raw_tensor=a.seq_lens))
        b_spatial_dim = Dim(Tensor("b_sizes", [batch_dim], dtype="int64", raw_tensor=b.seq_lens))
        a_ = Tensor("a", [batch_dim, a_spatial_dim], dtype="int64", raw_tensor=a.tensor)
        b_ = Tensor("b", [batch_dim, b_spatial_dim], dtype="int64", raw_tensor=b.tensor)
        _check_edit_distance(a_, a_spatial_dim, b_, b_spatial_dim)

    rnd = numpy.random.RandomState(42)
    for a, b in itertools.product(_SeqsB1, _SeqsB1):
        batch_size = rnd.randint(2, 11)
        a_max_len = rnd.randint(a.seq_lens[0], a.seq_lens[0] + 5)
        b_max_len = rnd.randint(b.seq_lens[0], b.seq_lens[0] + 5)
        a_sizes = rnd.randint(0, a_max_len + 1, size=(batch_size,))
        b_sizes = rnd.randint(0, b_max_len + 1, size=(batch_size,))
        a_sizes[0] = a.seq_lens[0]
        b_sizes[0] = b.seq_lens[0]
        a_max_len = max(a_sizes)
        b_max_len = max(b_sizes)
        a_values = rnd.randint(0, 10, (batch_size, a_max_len))
        b_values = rnd.randint(0, 10, (batch_size, b_max_len))
        a_values[0, : a.seq_lens[0]] = a.tensor[0, : a.seq_lens[0]]
        b_values[0, : b.seq_lens[0]] = b.tensor[0, : b.seq_lens[0]]
        a_sizes = torch.tensor(a_sizes, dtype=torch.int32)
        b_sizes = torch.tensor(b_sizes, dtype=torch.int32)

        # noinspection PyShadowingNames
        batch_dim = Dim(batch_size, name="batch")
        a_spatial_dim = Dim(Tensor("a_sizes", [batch_dim], dtype="int32", raw_tensor=a_sizes))
        b_spatial_dim = Dim(Tensor("b_sizes", [batch_dim], dtype="int32", raw_tensor=b_sizes))
        a_ = Tensor("a", [batch_dim, a_spatial_dim], dtype="int64", raw_tensor=torch.tensor(a_values))
        b_ = Tensor("b", [batch_dim, b_spatial_dim], dtype="int64", raw_tensor=torch.tensor(b_values))
        _check_edit_distance(a_, a_spatial_dim, b_, b_spatial_dim)


def test_audio_log_mel_filterbank_from_raw_bfloat16():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    out_dim = Dim(80, name="freq")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        audio = extern_data["data"]
        audio = rf.cast(audio, "bfloat16")
        out, out_spatial_dim = rf.audio.log_mel_filterbank_from_raw(audio, in_spatial_dim=time_dim, out_dim=out_dim)
        assert out.dtype == "bfloat16"
        out = rf.cast(out, "float32")  # the test framework doesn't support bfloat16 currently due to Numpy...
        out.mark_as_default_output(shape=(batch_dim, out_spatial_dim, out_dim))

    run_model(
        extern_data,
        lambda **_kwargs: rf.Module(),
        _forward_step,
        dyn_dim_min_sizes={time_dim: 2000},
        dyn_dim_max_sizes={time_dim: 3000},
        test_tensorflow=False,
    )


def test_copy_compatible_to_empty():
    import torch

    beam_dim = Dim(3, name="beam")
    hist_dim = Dim(0, name="hist")
    hist_bc_dim = Dim(1, name="hist_bc")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, beam_dim], dtype="int32"),
        }
    )
    out_template = Tensor("out_template", [batch_dim, beam_dim, hist_dim], dtype="int32")

    # noinspection PyShadowingNames
    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        data = extern_data["data"]

        out = data.copy_compatible_to(out_template)
        print("out:", out)
        assert (
            len(out.dims) == 3
            and out.dims[:2] == (batch_dim, beam_dim)
            and out.dims[2] != hist_dim
            and out.dims[2].dimension == 1
        )
        out, _ = rf.replace_dim(out, in_dim=out.dims[2], out_dim=hist_bc_dim)  # just that we have consistent output
        out.mark_as_output("out", shape=out.dims)

        out_unbroadcast = data.copy_compatible_to(out_template, unbroadcast=True)
        print("out_unbroadcast:", out_unbroadcast)
        assert out_unbroadcast.dims == out_template.dims
        if isinstance(out_unbroadcast.raw_tensor, torch.Tensor):
            print("out_unbroadcast.raw_tensor:", out_unbroadcast.raw_tensor)
            assert out_unbroadcast.raw_tensor.shape == (
                batch_dim.get_dim_value(),
                beam_dim.dimension,
                hist_dim.dimension,
            )
        out_unbroadcast.mark_as_output("out_unbroadcast", shape=out_unbroadcast.dims)

        out_raw = data.copy_compatible_to_dims_raw(out_template.dims)
        if isinstance(out_raw, torch.Tensor):
            print("out raw:", out_raw)
            assert out_raw.shape == out.raw_tensor.shape == (batch_dim.get_dim_value(), beam_dim.dimension, 1)
            torch.testing.assert_close(out_raw, out.raw_tensor)

    run_model(extern_data, lambda **_kwargs: rf.Module(), _forward_step, test_tensorflow=False)


def test_random_uniform_int():
    batch_dim = Dim(3, name="batch")
    feat_dim = Dim(5, name="feat")

    # noinspection PyShadowingNames
    def _forward_step(**_kwargs):
        out = rf.random_uniform([batch_dim, feat_dim], dtype="int32", minval=-10, maxval=10)
        out.mark_as_default_output(shape=[batch_dim, feat_dim])

    run_model(TensorDict(), lambda **_kwargs: rf.Module(), _forward_step)


def test_random_choice_with_replacement():
    batch_dim = Dim(3, name="batch")
    time_dim = Dim(7, name="time")
    feat_dim = Dim(5, name="feat")

    # noinspection PyShadowingNames
    def _forward_step(**_kwargs):
        probs = rf.softmax(rf.random_uniform([feat_dim], dtype="float32", minval=-1.0, maxval=1.0), axis=feat_dim)
        out = rf.random_choice_with_replacement([batch_dim], probs=probs, axis=feat_dim)
        assert out.dims == (batch_dim,) and out.sparse_dim == feat_dim
        out.mark_as_default_output(shape=[batch_dim])

        probs = rf.softmax(
            rf.random_uniform([batch_dim, feat_dim], dtype="float32", minval=-1.0, maxval=1.0), axis=feat_dim
        )
        out = rf.random_choice_with_replacement([batch_dim], probs=probs, axis=feat_dim)
        assert out.dims == (batch_dim,) and out.sparse_dim == feat_dim
        out.mark_as_output("batched_no_new", shape=[batch_dim])

        out = rf.random_choice_with_replacement([batch_dim, time_dim], probs=probs, axis=feat_dim)
        assert out.dims == (batch_dim, time_dim) and out.sparse_dim == feat_dim
        out.mark_as_output("batched_time", shape=[batch_dim, time_dim])

        probs = rf.softmax(
            rf.random_uniform([batch_dim, time_dim, feat_dim], dtype="float32", minval=-1.0, maxval=1.0), axis=feat_dim
        )
        out = rf.random_choice_with_replacement([batch_dim, time_dim], probs=probs, axis=feat_dim)
        assert out.dims == (batch_dim, time_dim) and out.sparse_dim == feat_dim
        out.mark_as_output("mult_common", shape=[batch_dim, time_dim])

    run_model(TensorDict(), lambda **_kwargs: rf.Module(), _forward_step, test_tensorflow=False)



"""
Provides :class:`PostprocessingDataset`.
"""

from __future__ import annotations

from collections import deque
from itertools import islice
import numpy
from numpy.random import RandomState
import select
import sys
import threading
from typing import Any, Callable, Dict, Iterator, List, Optional, Sequence, Tuple, TypeVar

from returnn.config import SubProcCopyGlobalConfigPreInitFunc
from returnn.datasets.basic import DatasetSeq
from returnn.datasets.util.strings import str_to_numpy_array
from returnn.datasets.util.vocabulary import Vocabulary
from returnn.tensor import Tensor, TensorDict
from returnn.tensor.dim import Dim
from returnn.util import basic as util, better_exchook
from returnn.util.multi_proc_non_daemonic_spawn import NonDaemonicSpawnContext
from .basic import Dataset, init_dataset
from .cached2 import CachedDataset2

# noinspection PyProtectedMember
from multiprocessing.connection import Connection as mpConnection

_mp = NonDaemonicSpawnContext(process_pre_init_func=SubProcCopyGlobalConfigPreInitFunc())


__all__ = ["PostprocessingDataset", "LaplaceOrdering", "Sequential"]


class PostprocessingDataset(CachedDataset2):
    """
    A dataset that allows for generic online post-processing of data from another
    dataset using a function on the segment level and on the level of multiple
    segments via an iterator.

    This allows integrating various data augmentation techniques like e.g. Mixup,
    SpecAugment or speed perturbation into the data loading pipeline.

    The integration into the data loading pipeline makes it easy to distribute the
    data processing work across multiple CPU cores and in turn frees the GPU from
    data preprocessing tasks.

    Multiprocessing can either be done using :class:``MultiProcDataset`` or by setting
    `num_workers > 0` on this class.

    The latter only applies parallelism to the post-processing functions themselves,
    and does not duplicate the underlying dataset once per worker.
    This is often fast enough and has the advantage of lower memory consumption.

    Example usage::

        from returnn.tensor.dim import Dim, DimTypes

        time_dim = Dim(None, kind=DimTypes.Spatial)
        new_data_dim = Dim(128)

        train = {
            "class": "PostprocessingDataset",
            "dataset": {
                "class": "HDFDataset",
                "files": ["/path/to/data.hdf"],
            },
            # one of them, but not both:
            # (data: TensorDict, *, rng: numpy.random.RandomState, **kwargs) -> TensorDict
            "map_seq": map_seq,
            # (iter: Iterator[TensorDict], *, rng: numpy.random.RandomState, **kwargs) -> Iterator[TensorDict]
            "map_seq_stream": map_seqs,
            # only required when data shapes change wrt. the wrapped dataset:
            "map_outputs": {
                "data": {"dims": [time_dim, new_data_dim]},
            },
        }

    The postprocessor functions operate on ``TensorDict``s, which have entries for
    all data keys in the underlying dataset.

    There may also be additional "meta" entries in the tensor dicts, like ``complete_frac``,
    ``seq_idx`` and ``seq_tag``.
    These should be copied over in a manner that is reasonable for the use case at hand and
    ensures forwards compatibility as well as reasonably possible.

    The dataset itself does not support its own seq ordering and relies on the wrapped
    dataset for seq ordering instead. Specifying a ``seq_ordering`` other than ``default``
    results in an error.

    However, we provide an iterator that implements the common `laplace:.NUM_SEQS_PER_BIN`-variant
    of seq ordering that any custom ``map_seq_stream``-style postprocessing iterator can be composed
    with to implement the ordering via :class:`LaplaceOrdering`.

    Like this::

        from returnn.datasets.postprocessing import LaplaceOrdering, Sequential

        def my_map_seq_stream(iterator):
            ...

        train = {
            "class": "PostprocessingDataset",
            # ...
            "map_seq_stream": Sequential(
                my_map_seq_stream,
                LaplaceOrdering(num_seqs_per_bin=1000),
            ),
        }
    """

    def __init__(
        self,
        *,
        dataset: Dict[str, Any],
        map_seq: Optional[Callable] = None,
        map_seq_stream: Optional[Callable] = None,
        map_outputs: Optional[Dict[str, Any]] = None,
        map_seq_stream_preserves_num_seqs: Optional[bool] = None,
        buf_size: int = 1,
        num_workers: int = 0,
        **kwargs,
    ):
        """
        :param dataset: inner dataset to be post-processed
        :param map_seq: post processor function operating on the single-segment level.
            Signature: `(data: TensorDict, *, rng: numpy.random.RandomState, **kwargs) -> TensorDict`
            To avoid confusion on the order of how the processing functions are applied to the data, only one of
            ``map_seq`` and ``map_seq_stream`` can be specified at a time.
            To ensure forwards compatibility, the function must accept ``**kwargs`` as its last argument.
            This is enforced by passing randomly named parameters at runtime.
        :param map_seq_stream: post processor function operating on the multiple segment level via an iterator.
            Allows merging multiple segments into one, or generating multiple output segments from one input segment.
            Signature:
                ``(iter: Iterator[TensorDict], *, rng: numpy.random.RandomState, **kwargs) -> Iterator[TensorDict]``
            To avoid confusion on the order of how the processing functions are applied to the data, only one of
            ``map_seq`` and ``map_seq_stream`` can be specified at a time.
            To ensure forwards compatibility, the function must accept ``**kwargs`` as its last argument.
            This is enforced by passing randomly named parameters at runtime.
        :param map_outputs: Type and axis specification of the outputs of the mapping functions,
            like extern_data and model_outputs.
            To simplify the common case when no shapes change, this value can be left unspecified. The dataset then
            assumes the same data layout as returned by the wrapped dataset.
            Example: `map_outputs={"data": {"dim": 42}}`
        :param map_seq_stream_preserves_num_seqs: whether the function in map_seq_stream preserves the number of
            sequences, i.e. for every input sequence there is exactly one output sequence.
        :param buf_size: Buffer size for each worker, number of seqs to prefetch. Must be > 0.
        :param num_workers: If > 0, configures the number of worker processes to use for data postprocessing.
            Only the postprocessing is distributed across subprocesses,
            the underlying dataset is only instantiated once.
            This usually has lower memory consumption than using :class:``MultiProcDataset``.
        :param kwargs: see :class:`CachedDataset2`, :class:`Dataset`
        """
        super().__init__(**kwargs)

        if self.seq_ordering != "default":
            raise ValueError(f"{self}: specify seq_ordering in wrapped dataset, not in {self.__class__.__name__}")
        if map_seq is None and map_seq_stream is None:
            raise ValueError(f"{self}: need to either set map_seq or map_seq_stream")
        if map_seq and map_seq_stream:
            raise ValueError(f"{self}: cannot set both map_seq and map_seq_stream")
        if map_seq and map_seq_stream_preserves_num_seqs is not None:
            raise ValueError(f"{self}: map_seq_stream_preserves_num_seqs is only allowed with map_seq_stream")

        if buf_size < 1:
            raise ValueError(f"{self}: buf_size must be > 0, but got {buf_size}")
        if num_workers < 0:
            raise ValueError(f"{self}: num_workers must be >= 0, but got {num_workers}")

        self._dataset_def = dataset
        self._map_seq = map_seq
        self._map_seq_stream = map_seq_stream
        if map_seq_stream_preserves_num_seqs is None and map_seq_stream is not None:
            map_seq_stream_preserves_num_seqs = getattr(map_seq_stream, "preserves_num_seqs", None)
        assert map_seq_stream_preserves_num_seqs is None or isinstance(map_seq_stream_preserves_num_seqs, bool)
        self._map_seq_stream_preserves_num_seqs = map_seq_stream_preserves_num_seqs
        self._map_outputs = map_outputs
        self._seq_list_for_validation: Optional[List[str]] = None

        self._dataset = init_dataset(self._dataset_def, parent_dataset=self)
        if self._map_seq_stream is None or self._map_seq_stream_preserves_num_seqs is True:
            # if the stream mapper is set, the num_seqs may change and the estimation is less accurate
            self._estimated_num_seqs = self._dataset.estimated_num_seqs
        self._data_iter: Optional[Iterator[Tuple[int, TensorDict]]] = None
        self._data_iter_produced_num_seqs = 0

        self._buf_size = buf_size
        # Ensure only one feeder thread at a time accesses the wrapped dataset to
        # prevent race conditions while moving from one epoch to the next.
        self._dataset_lock = threading.Lock()
        self._multi_proc_data_iter: Optional[_MultiProcDataIter] = None  # store for cleanup
        self._num_workers = num_workers
        self._worker_procs: Optional[List[_WorkerProcParent]] = None

        self._in_tensor_dict_template = TensorDict(
            {name: self._make_tensor_template_from_input(name) for name in self._dataset.get_data_keys()}
        )
        self.labels = {}
        if self._map_outputs is not None:
            self._out_tensor_dict_template = TensorDict()
            self._out_tensor_dict_template.update(self._map_outputs, auto_convert=True)
        else:
            self._out_tensor_dict_template = self._in_tensor_dict_template.copy_template()
            self.labels = self._dataset.labels.copy()
        # update only after _out_tensor_dict_template has been created from _in_tensor_dict_template
        self._in_tensor_dict_template.update(
            {
                "complete_frac": {"dims": (), "dtype": "float32"},
                "seq_idx": {"dims": (), "dtype": "int32"},
                "seq_tag": {"dims": (), "dtype": "string"},
            },
            auto_convert=True,
        )
        self.num_outputs = {
            k: (t.sparse_dim.size if t.sparse_dim else t.shape[-1] if len(t.shape) > 0 else 1, t.ndim)
            for k, t in self._out_tensor_dict_template.data.items()
        }
        self._default_input = "data" if "data" in self.num_outputs else next(iter(self.num_outputs.keys()))
        self.num_inputs = self.num_outputs[self._default_input][0]

        for k, t in self._out_tensor_dict_template.data.items():
            if self.labels.get(k):
                continue
            if t.vocab:
                self.labels[k] = t.vocab.labels
            elif t.sparse_dim:  # sparse_dim but not vocab
                self.labels[k] = list(map(str, range(t.sparse_dim.dimension)))  # dummy labels

    def init_seq_order(
        self, epoch: Optional[int] = None, seq_list: Optional[List[str]] = None, seq_order: Optional[List[int]] = None
    ):
        """
        :param epoch:
        :param seq_list:
        :param seq_order:
        :return: whether the order changed (True is always safe to return)
        """
        super().init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)

        if self._map_seq_stream is not None:
            if seq_list is not None:
                raise ValueError("map_seq_stream is set, cannot specify custom seq_list")
            if seq_order is not None:
                raise ValueError("map_seq_stream is set, cannot specify custom seq_order")

        if self._multi_proc_data_iter is not None:
            self._multi_proc_data_iter.stop()
            self._multi_proc_data_iter = None

        if epoch is None and seq_list is None and seq_order is None:
            self._num_seqs = 0
            return True

        if self._num_workers > 0:
            self._lazy_init_worker_procs()
            assert self._worker_procs is not None and len(self._worker_procs) == self._num_workers
            parent_conns, child_conns = zip(*[_mp.Pipe() for _ in range(self._num_workers)])
            base_rng_seed = self._get_random_seed_for_epoch(epoch=epoch) * 683859 * self._num_workers
            for i, (worker, child_conn) in enumerate(zip(self._worker_procs, child_conns)):
                worker.init_seq_order(
                    epoch=epoch,
                    rng_seed=(base_rng_seed + 30411 * i) % (2**32 - 1),
                    seq_list=seq_list,
                    seq_pipe=child_conn,
                )
            data_iter = self._multi_proc_data_iter = self._init_multi_proc_data_iter(
                epoch=epoch, feeder_to_worker_conns=parent_conns, seq_list=seq_list, seq_order=seq_order
            )
        else:
            self._dataset.init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)
            data_iter = _build_mapping_iter(
                _iterate_dataset(self._dataset, in_tensor_dict_template=self._in_tensor_dict_template),
                map_seq=self._map_seq,
                map_seq_stream=self._map_seq_stream,
                epoch=epoch,
                out_tensor_dict_template=self._out_tensor_dict_template,
                rng=RandomState(self._get_random_seed_for_epoch(epoch=epoch)),
                seq_list_for_validation=seq_list,
            )
        self._data_iter = enumerate(data_iter)
        self._data_iter_produced_num_seqs = 0
        self._seq_list_for_validation = seq_list
        if self._map_seq_stream is None or self._map_seq_stream_preserves_num_seqs is True:
            # If we don't have an iterable mapper (or the user explicitly specifies this),
            # we know the number of segments exactly equals the number of segments in the wrapped dataset
            try:
                self._num_seqs = self._dataset.num_seqs
            except NotImplementedError:
                pass  # some datasets don't know their num_seqs
        return True

    def __del__(self):
        if self._multi_proc_data_iter is not None:
            self._multi_proc_data_iter.stop(join=True)
            self._multi_proc_data_iter = None
        if not self._worker_procs:
            return
        got_exception = False
        for parent in self._worker_procs:
            # noinspection PyBroadException
            try:
                parent.exit(join=False)
            except Exception:
                got_exception = True
        if got_exception:
            return
        for parent in self._worker_procs:
            util.try_run(parent.worker_proc.join)

    def get_current_seq_order(self):
        """:return: current seq order of wrapped dataset, if map_seq_stream is not used"""
        if self._map_seq_stream is not None:
            raise Exception(f"{self}: get_current_seq_order is not allowed when map_seq_stream is set.")
        assert self._dataset is not None
        return self._dataset.get_current_seq_order()

    def get_data_keys(self):
        """:return: available data keys"""
        return list(self._out_tensor_dict_template.data.keys())

    def get_data_dtype(self, key):
        """:return: dtype of data entry `key`"""
        return self._out_tensor_dict_template.data[key].dtype

    def get_total_num_seqs(self, *, fast=False):
        """:return: total num seqs excluding partition_epoch"""
        if self._map_seq_stream is not None:
            raise util.OptionalNotImplementedError(
                f"{self}: get_total_num_seqs not allowed when map_seq_stream is set."
            )
        assert self._dataset is not None
        return self._dataset.get_total_num_seqs(fast=fast)

    def get_all_tags(self) -> List[str]:
        """:return: all tags"""
        if self._map_seq_stream is not None:
            raise util.OptionalNotImplementedError(f"{self}: get_all_tags not allowed when map_seq_stream is set.")
        assert self._dataset is not None
        return self._dataset.get_all_tags()

    def supports_sharding(self) -> bool:
        """:return: whether this dataset supports sharding"""
        assert self._dataset is not None
        return self._dataset.supports_sharding()

    def finish_epoch(self, *, free_resources=False):
        """finish_epoch"""
        super().finish_epoch(free_resources=free_resources)
        if not free_resources:
            return
        if self._multi_proc_data_iter is not None:
            self._multi_proc_data_iter.stop(join=True)
            self._multi_proc_data_iter = None
        if self._worker_procs is not None:
            for wp in self._worker_procs:
                wp.exit(join=True)
            self._worker_procs = None

    def _collect_single_seq(self, seq_idx: int) -> Optional[DatasetSeq]:
        while True:
            try:
                loaded_seq_idx, tensor_dict = next(self._data_iter)
                self._data_iter_produced_num_seqs += 1
                if self._num_seqs is not None:
                    assert self._data_iter_produced_num_seqs <= self._num_seqs, (
                        f"{self}: map_seq_stream yielded more seqs ({self._data_iter_produced_num_seqs}) "
                        f"than expected ({self._num_seqs}). map_seq_stream_preserves_num_seqs is set to "
                        f"{self._map_seq_stream_preserves_num_seqs}"
                    )
            except StopIteration:
                if self._num_seqs is not None:
                    assert self._data_iter_produced_num_seqs == self._num_seqs, (
                        f"{self}: map_seq_stream yielded {self._data_iter_produced_num_seqs} seqs, "
                        f"while {self._num_seqs} were expected. map_seq_stream_preserves_num_seqs is set to "
                        f"{self._map_seq_stream_preserves_num_seqs}"
                    )
                return None
            assert loaded_seq_idx <= seq_idx, "_collect_single_seq must be done monotonically"
            if loaded_seq_idx != seq_idx:
                continue
            complete_frac = (
                float(tensor_dict.data["complete_frac"].raw_tensor) if "complete_frac" in tensor_dict.data else None
            )
            seq_tag = str(tensor_dict.data["seq_tag"].raw_tensor) if "seq_tag" in tensor_dict.data else f"seq-{seq_idx}"
            features = {k: t.raw_tensor for k, t in tensor_dict.data.items() if k not in ["complete_frac", "seq_tag"]}
            seq = DatasetSeq(complete_frac=complete_frac, features=features, seq_idx=seq_idx, seq_tag=seq_tag)
            return seq

    def _make_tensor_template_from_input(self, data_key: str) -> Tensor:
        dtype = self._dataset.get_data_dtype(data_key)
        if dtype == "string":
            dims = []
        else:
            feature_dims = [
                Dim(dimension=dim, name=f"{data_key}_dim{i + 1}")
                for i, dim in enumerate(self._dataset.get_data_shape(data_key))
            ]
            dims = [Dim(dimension=None, name=f"{data_key}_frame"), *feature_dims]
        sparse_dim = None
        if self._dataset.is_data_sparse(data_key):
            sparse_dim = Dim(dimension=self._dataset.get_data_dim(data_key), name=f"{data_key}_sparse")
            labels = self._dataset.labels.get(data_key)
            if labels and len(labels) == sparse_dim.dimension:
                sparse_dim.vocab = Vocabulary.create_vocab_from_labels(self._dataset.labels[data_key])
        return Tensor(data_key, dims=dims, dtype=dtype, sparse_dim=sparse_dim)

    def _lazy_init_worker_procs(self):
        if self._worker_procs is not None:
            return
        self._worker_procs = [
            _WorkerProcParent(
                name=f"{self.__class__.__name__} {self.name} worker",
                buffer_size=self._buf_size,
                index=i,
                map_seq=self._map_seq,
                map_seq_stream=self._map_seq_stream,
                out_tensor_dict_template=self._out_tensor_dict_template,
            )
            for i in range(self._num_workers)
        ]

    def _init_multi_proc_data_iter(
        self,
        *,
        epoch: int,
        feeder_to_worker_conns: Sequence[mpConnection],
        seq_list: Optional[List[str]] = None,
        seq_order: Optional[List[int]] = None,
    ) -> _MultiProcDataIter:
        assert len(feeder_to_worker_conns) == self._num_workers

        quit_event = threading.Event()
        dataset_thread = threading.Thread(
            target=self._init_seq_order_and_distribute_seqs_to_children,
            kwargs={
                "epoch": epoch,
                "quit_event": quit_event,
                "seq_list": seq_list,
                "seq_order": seq_order,
                "worker_conns": feeder_to_worker_conns,
            },
            name=f"{self.__class__.__name__} feeder ep {epoch}",
        )
        # parent_conns are not closed here, because they move to a different thread, not process,
        # and so they must remain open.
        dataset_thread.start()
        data_iter = _MultiProcDataIter(
            dataset_thread=dataset_thread, quit_event=quit_event, worker_procs=self._worker_procs
        )
        return data_iter

    def _init_seq_order_and_distribute_seqs_to_children(
        self,
        *,
        epoch: int,
        quit_event: threading.Event,
        seq_list: Optional[List[str]] = None,
        seq_order: Optional[List[int]] = None,
        worker_conns: Sequence[mpConnection],
    ):
        """
        Initialize the wrapped dataset and distribute the contained sequences to the child worker processes.
        """

        assert self._buf_size > 0
        assert len(worker_conns) > 0
        assert self._num_workers > 0

        caches: List[deque[TensorDict]] = [deque() for _ in range(len(worker_conns))]

        def _any_conn_ready() -> bool:
            ready, _, _ = select.select(worker_conns, [], [], 0)
            return len(ready) > 0

        def _maybe_distrib_seq(*, timeout=0.1):
            assert timeout >= 0.0
            # do not block indefinitely to periodically check the quit_event
            ready_conns, _, _ = select.select(worker_conns, [], [], timeout)
            assert len(worker_conns) == len(caches)
            for child_queue, cache in zip(worker_conns, caches):
                if child_queue not in ready_conns:
                    continue
                msg, _ = child_queue.recv()
                assert msg == "get_seq"
                tensor_dict = cache.popleft() if len(cache) > 0 else None
                child_queue.send(("seq", tensor_dict))

        # Lock ensures that only one thread at a time accesses the wrapped dataset.
        # This protects against issues while moving from one epoch to the next.
        with self._dataset_lock:
            self._dataset.init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)
            data_iter = _iterate_dataset(self._dataset, in_tensor_dict_template=self._in_tensor_dict_template)
            data_iter = enumerate(data_iter)

            def _add_to_cache() -> bool:
                try:
                    idx, tensor_dict = next(data_iter)
                    caches[idx % len(caches)].append(tensor_dict)
                    return True
                except StopIteration:
                    return False

            while not quit_event.is_set():
                # fetch seqs until all caches have at least one seq,
                # if no child is waiting for seqs also fill until buf_size
                while any(len(cache) == 0 for cache in caches) or (
                    sum(len(cache) for cache in caches) < self._buf_size and not _any_conn_ready()
                ):
                    if not _add_to_cache():
                        break
                if all(len(c) == 0 for c in caches):
                    break
                try:
                    _maybe_distrib_seq()
                except (BrokenPipeError, EOFError):
                    # queue is closed, i.e. the worker process crashed for some reason -> stop
                    break

        for queue in worker_conns:
            try:
                queue.send(("seq", None))
            except (BrokenPipeError, EOFError):
                # queue is already closed, i.e. the worker process died
                pass
            finally:
                queue.close()


def _iterate_dataset(dataset: Dataset, *, in_tensor_dict_template: TensorDict) -> Iterator[TensorDict]:
    """
    :return: generator providing data samples in the form of a TensorDict
    """
    data_keys = dataset.get_data_keys()

    seq_index = 0
    while dataset.is_less_than_num_seqs(seq_index):
        dataset.load_seqs(seq_index, seq_index + 1)

        tensor_dict = in_tensor_dict_template.copy_template()
        for data_key in data_keys:
            tensor_dict.data[data_key].raw_tensor = dataset.get_data(seq_index, data_key)

        complete_frac = dataset.get_complete_frac(seq_index, allow_only_lr_suitable=True)
        if complete_frac is not None:
            comp_frac_raw_tensor = numpy.array(complete_frac, dtype=numpy.float32)
            tensor_dict.data["complete_frac"].raw_tensor = comp_frac_raw_tensor
        seq_idx_raw_tensor = numpy.array(seq_index, dtype=numpy.int32)
        tensor_dict.data["seq_idx"].raw_tensor = seq_idx_raw_tensor
        seq_tag_raw_tensor = str_to_numpy_array(dataset.get_tag(seq_index))
        tensor_dict.data["seq_tag"].raw_tensor = seq_tag_raw_tensor

        yield tensor_dict
        seq_index += 1


def _build_mapping_iter(
    data_iter: Iterator[TensorDict],
    *,
    map_seq: Optional[Callable] = None,
    map_seq_stream: Optional[Callable] = None,
    epoch: int,
    out_tensor_dict_template: TensorDict,
    rng: RandomState,
    seq_list_for_validation: Optional[List[str]] = None,
) -> Iterator[TensorDict]:
    """
    Build an iterator applying the mapping functions on the given dataset iterator.

    :param data_iter: iterator providing data samples in the form of a TensorDict
    :param map_seq: see :class:`PostprocessingDataset`
    :param map_seq_stream: see :class:`PostprocessingDataset`
    :param epoch: current epoch number
    :param out_tensor_dict_template: template for the output TensorDicts, used for validation
    :param rng: random number generator to use
    :param seq_list_for_validation: optional list of seq tags to validate against when processing the data
    :return: an iterator applying both the segment level and across-segment transformations on the given dataset
    """

    def _validate_tensor_dict_iter(inner: Iterator[TensorDict]) -> Iterator[TensorDict]:
        last_complete_frac = 0.0
        for t_dict in inner:
            assert isinstance(t_dict, TensorDict), (
                f"postprocessing mapper function must produce a {TensorDict.__name__}, "
                f"but got a {type(t_dict).__name__}"
            )
            if "complete_frac" in t_dict.data:  # sanity check complete_frac
                complete_frac = float(t_dict.data["complete_frac"].raw_tensor)
                assert 0.0 <= complete_frac <= 1.0, f"complete_frac must be in [0, 1], but got {complete_frac}"
                assert complete_frac >= last_complete_frac, (
                    "complete_frac must be monotonically increasing, "
                    f"but got {complete_frac} after {last_complete_frac}"
                )
                last_complete_frac = complete_frac
            for data_key, out_t in out_tensor_dict_template.data.items():
                in_t = t_dict.data[data_key]
                assert in_t.ndim == out_t.batch_ndim, (
                    f"Dim number mismatch for {data_key}: {in_t.ndim} != {out_t.batch_ndim}. "
                    "Postprocessing data tensors must not have a batch dimension."
                )
                assert in_t.dtype == out_t.dtype, f"dtype mismatch for {data_key}: '{in_t.dtype}' != '{out_t.dtype}'"
                for i, (in_dim, out_shape) in enumerate(zip(in_t.dims, out_t.shape)):
                    assert in_dim.dimension is None or in_dim.dimension == out_shape, (
                        f"Dim {i} mismatch on {data_key}: {in_dim.dimension} must either be `None` or equal {out_shape}"
                    )
            yield t_dict

    def _apply_map_seq(tensor_dict: TensorDict) -> TensorDict:
        comp_frac_raw_tensor = (
            tensor_dict.data["complete_frac"].raw_tensor if "complete_frac" in tensor_dict.data else None
        )
        seq_index_raw = tensor_dict.data["seq_idx"].raw_tensor
        seq_index = int(seq_index_raw.item())
        seq_tag_raw_tensor = tensor_dict.data["seq_tag"].raw_tensor

        tensor_dict = map_seq(tensor_dict, epoch=epoch, seq_idx=seq_index, rng=rng, **util.get_fwd_compat_kwargs())
        assert isinstance(tensor_dict, TensorDict), (
            f"map_seq must produce a {TensorDict.__name__}, but produced {type(tensor_dict).__name__}"
        )

        # Re-adding the complete_frac/seq_idx/seq_tag here causes no harm in case they are dropped
        # since we don't add/drop any segments w/ the non-iterator postprocessing function.
        if "complete_frac" not in tensor_dict.data and comp_frac_raw_tensor is not None:
            tensor_dict.data["complete_frac"] = Tensor(
                "complete_frac", dims=(), dtype="float32", raw_tensor=comp_frac_raw_tensor
            )
        if "seq_idx" not in tensor_dict.data:
            tensor_dict.data["seq_idx"] = Tensor("seq_idx", dims=(), dtype="int32", raw_tensor=seq_index_raw)
        if "seq_tag" not in tensor_dict.data:
            tensor_dict.data["seq_tag"] = Tensor("seq_tag", dims=(), dtype="string", raw_tensor=seq_tag_raw_tensor)

        if seq_list_for_validation is not None:
            seq_tag = seq_list_for_validation[seq_index]
            tag_of_seq = tensor_dict.data["seq_tag"].raw_tensor.item()
            assert tag_of_seq == seq_tag, (
                f"seq tag mismath: {tag_of_seq} != {seq_tag} for seq index {seq_index} when seq list is given"
            )

        return tensor_dict

    assert map_seq or map_seq_stream, "need to specify either map_seq or map_seq_stream"
    assert not (map_seq and map_seq_stream), "cannot set both map_seq and map_seq_stream"
    if map_seq is not None:
        data_iter = (_apply_map_seq(t_dict) for t_dict in data_iter)
    if map_seq_stream is not None:
        data_iter = map_seq_stream(data_iter, epoch=epoch, rng=rng, **util.get_fwd_compat_kwargs())
        assert isinstance(data_iter, Iterator), (
            f"map_seq_stream must produce an {Iterator.__name__}, but produced {type(data_iter).__name__}"
        )
    return _validate_tensor_dict_iter(data_iter)


class _MultiProcDataIter:
    """
    Data iter that pulls from the worker processes in a well-defined order and
    manages the lifetime of the feeder thread.

    Also ensures monotonicity of complete_frac, which would otherwise be no longer
    guaranteed if there is more than one worker.
    """

    def __init__(
        self, *, dataset_thread: threading.Thread, quit_event: threading.Event, worker_procs: List[_WorkerProcParent]
    ):
        self.dataset_thread = dataset_thread
        self.quit_event = quit_event
        assert len(worker_procs) > 0
        self.worker_procs = worker_procs

        self._complete_frac = 0.0  # need to force monotonicity
        self._workers_exhausted = [False for _ in range(len(worker_procs))]
        self._worker_idx = 0

    def __iter__(self):
        return self

    def __next__(self) -> Optional[TensorDict]:
        if self.quit_event.is_set():
            raise StopIteration

        while True:
            if all(self._workers_exhausted):
                break
            worker_idx = self._worker_idx
            self._worker_idx = (self._worker_idx + 1) % len(self.worker_procs)
            if self._workers_exhausted[worker_idx]:
                continue
            seq = self.worker_procs[worker_idx].get_seq()
            if seq is not None:
                return self._ensure_complete_frac_monotonic(seq)
            self._workers_exhausted[worker_idx] = True

        # when we reach this point, all workers are exhausted and we stop
        self.stop()
        raise StopIteration

    def stop(self, *, join=True):
        """
        Stop the iterator and the dataset thread.

        Once this is called, the iterator cannot be used anymore.
        """
        if self.quit_event.is_set():
            return
        self.quit_event.set()
        if join:
            util.try_run(self.dataset_thread.join)

    def _ensure_complete_frac_monotonic(self, seq: TensorDict) -> TensorDict:
        """
        Enforce monotonicity of `complete_frac` in the given `TensorDict`.
        """
        if "complete_frac" not in seq.data:
            return seq
        complete_frac = float(seq.data["complete_frac"].raw_tensor)
        assert 0.0 <= complete_frac <= 1.0, f"complete_frac must be in [0, 1], but got {complete_frac}"
        self._complete_frac = max(complete_frac, self._complete_frac)
        seq.data["complete_frac"].raw_tensor = numpy.array(self._complete_frac, dtype=numpy.float32)
        return seq

    def __del__(self):
        # noinspection PyBroadException
        try:
            self.stop(join=False)
        except Exception:
            pass


class _WorkerProcParent:
    def __init__(
        self,
        *,
        buffer_size: int,
        index: int,
        name: str,
        map_seq: Optional[Callable],
        map_seq_stream: Optional[Callable],
        out_tensor_dict_template: TensorDict,
    ):
        parent_conn, child_conn = _mp.Pipe()
        self.parent_conn = parent_conn

        self.worker_proc = _mp.Process(
            name=f"{name} worker {index}",
            target=_worker_proc_loop,
            args=(index, child_conn, buffer_size, map_seq, map_seq_stream, out_tensor_dict_template),
            daemon=True,
        )
        self.worker_proc.start()

        # Make sure the child connection is closed here.
        # It stays open in the child, until the child dies.
        # When that happens, now any consecutive read on the pipe
        # should yield an exception -- which is what we want,
        # otherwise it would just hang.
        child_conn.close()

    def init_seq_order(
        self,
        *,
        epoch: int,
        rng_seed: int,
        seq_list: Optional[List[str]],
        seq_pipe: mpConnection,
    ):
        """init_seq_order"""
        args = {"epoch": epoch, "rng_seed": rng_seed, "seq_list": seq_list, "seq_pipe": seq_pipe}
        self.parent_conn.send(("init_seq_order", args))
        msg, _ = self.parent_conn.recv()
        assert msg == "init_seq_order"
        # seq_pipe is owned by the child process,
        # and so must be closed in the parent to avoid hangs
        seq_pipe.close()

    def get_seq(self) -> Optional[TensorDict]:
        """get_seq"""
        self.parent_conn.send(("get_seq", {}))
        msg, seq = self.parent_conn.recv()
        assert msg == "seq"
        return seq

    def exit(self, *, join: bool = True):
        """exit"""
        self.parent_conn.send(("exit", {}))
        if join:
            self.worker_proc.join()

    def __del__(self):
        # noinspection PyBroadException
        try:
            self.exit(join=False)
        except Exception:
            pass
        else:
            util.try_run(self.worker_proc.join)


def _worker_proc_loop(
    index: int,
    parent_conn: mpConnection,
    buffer_size: int,
    map_seq: Optional[Callable],
    map_seq_stream: Optional[Callable],
    out_tensor_dict_template: TensorDict,
):
    if sys.platform == "linux":
        with open("/proc/self/comm", "w") as f:
            f.write(f"PP worker {index}")
    better_exchook.setup_all()

    assert isinstance(buffer_size, int) and buffer_size > 0
    assert isinstance(index, int)
    assert isinstance(parent_conn, mpConnection)

    cache: deque[TensorDict] = deque()

    data_iter: Optional[Iterator[TensorDict]] = None
    feeder_conn: Optional[mpConnection] = None

    def _add_to_cache():
        nonlocal data_iter
        if data_iter is None:
            return False
        try:
            seq = next(data_iter)
        except StopIteration:
            data_iter = None
            return False
        cache.append(seq)
        return True

    def _iter_pipe(q: mpConnection) -> Iterator[TensorDict]:
        assert isinstance(q, mpConnection)

        while True:
            try:
                q.send(("get_seq", None))
                seq_msg, item = q.recv()
            except (BrokenPipeError, EOFError):
                # queue is closed
                break
            assert seq_msg == "seq"
            if item is None:
                break
            assert isinstance(item, TensorDict)
            yield item

    try:
        while True:
            while len(cache) < buffer_size and not parent_conn.poll():
                if not _add_to_cache():
                    break
            msg, kwargs = parent_conn.recv()
            if msg == "exit":
                break
            elif msg == "get_seq":
                if not cache:
                    _add_to_cache()
                parent_conn.send(("seq", cache.popleft() if cache else None))
            elif msg == "init_seq_order":
                epoch = kwargs["epoch"]
                if sys.platform == "linux":
                    with open("/proc/self/comm", "w") as f:
                        f.write(f"PP worker {index} ep {epoch}")
                if feeder_conn is not None:
                    feeder_conn.close()
                feeder_conn = kwargs["seq_pipe"]
                data_iter = _build_mapping_iter(
                    _iter_pipe(feeder_conn),
                    epoch=epoch,
                    map_seq=map_seq,
                    map_seq_stream=map_seq_stream,
                    out_tensor_dict_template=out_tensor_dict_template,
                    rng=RandomState(kwargs["rng_seed"]),
                    seq_list_for_validation=kwargs["seq_list"],
                )
                assert isinstance(data_iter, Iterator)
                cache.clear()
                parent_conn.send(("init_seq_order", None))
            else:
                raise Exception(f"unknown msg {msg!r}")
    except KeyboardInterrupt:  # when parent dies
        pass
    except EOFError:  # when parent dies
        pass
    finally:
        if feeder_conn is not None:
            feeder_conn.close()
        parent_conn.close()


class LaplaceOrdering(Callable[[Iterator[TensorDict]], Iterator[TensorDict]]):
    """
    Iterator compatible with :class:`PostprocessingDataset`'s ``map_seq_stream`` applying
    laplace sequence ordering based on the number of segments per bin.

    To be composed with any custom data postprocessing logic via :class:`Sequential`.
    """

    preserves_num_seqs = True

    def __init__(self, num_seqs_per_bin: int, length_key: str = "data"):
        """
        :param num_seqs_per_bin: number of segments in a single laplace bin.
        :param length_key: data key to determine the segment length from for ordering.
        """
        self.length_key = length_key
        assert num_seqs_per_bin > 0
        self.num_seqs_per_bin = num_seqs_per_bin

    def __call__(self, iterator: Iterator[TensorDict], **kwargs) -> Iterator[TensorDict]:
        """:return: generator applying laplace sequence ordering on the data"""
        iterator = iter(iterator)
        is_down_phase = False

        seq_buffer = list(islice(iterator, self.num_seqs_per_bin))
        has_ended = False
        while True:
            # Make sure to not reorder the monotonically increasing values for complete_frac
            # so that the trainer can calculate the appropriate learning rates.
            complete_frac_values = [tdict.data["complete_frac"].raw_tensor for tdict in seq_buffer]
            seq_buffer.sort(key=self._get_seq_len, reverse=is_down_phase)
            for sorted_item, comp_frac in zip(seq_buffer, complete_frac_values):
                sorted_item.data["complete_frac"].raw_tensor = comp_frac

            next_seq_buffer = []

            # Yield items to trainer while gradually pulling more data from PP function.
            # This optimizes CPU load when multiple workers are used.
            for item in seq_buffer:
                yield item

                try:
                    if not has_ended:
                        next_seq_buffer.append(next(iterator))
                except StopIteration:
                    has_ended = True

            if len(seq_buffer) < self.num_seqs_per_bin:
                assert has_ended and not next_seq_buffer
                break

            is_down_phase = not is_down_phase
            seq_buffer = next_seq_buffer

    def _get_seq_len(self, tdict: TensorDict) -> int:
        """
        :return: segment length of the segment in `tdict` as measured by `self.length_key` for comparison.
        """
        return tdict.data[self.length_key].raw_tensor.shape[0]


T = TypeVar("T", TensorDict, Iterator[TensorDict])


class Sequential:
    """
    Callable that composes multiple postprocessing functions into one by sequential application,
    i.e. Sequential(f, g)(x) = (g ∘ f)(x) = g(f(x)).

    Can either compose ``map_seq``-style single-segment processor functions or ``map_seq_stream``-style
    iterators operating on multiple segments. Just make sure not to mix both styles.
    """

    def __init__(self, *postprocessing_funcs: Callable):
        """
        :param postprocessing_funcs: Postprocessing functions to compose.
        """
        self.funcs = postprocessing_funcs

    def __call__(self, arg: T, **kwargs) -> T:
        """:return: result of sequential application of the postprocessing functions"""

        for func in self.funcs:
            arg = func(arg, **kwargs)
        return arg

    @property
    def preserves_num_seqs(self):
        """:return: whether the composed functions all preserve the number of sequences"""
        return all(getattr(f, "preserves_num_seqs", False) for f in self.funcs)
